
from .git_repository import GitRepository
from .box import Box
from .exceptions import ProjectException

class Project(object):
    def __init__(self, name, description = '', repositories = []):
        self.name = name
        self.description = description
        self.repositories = repositories
        self.db = None

    def news(self, fromcache):
        for name in self.repositories:
            repo = self.db.repositories[name]
            if not fromcache:
                repo.fetch()
            commits = repo.get_new_commits()
            if len(commits):
                yield Box(repo.name, "\n".join(commits))

    def diff(self):
        for name in self.repositories:
            repo = self.db.repositories[name]
            diff = repo.diff()
            if len(diff):
                yield Box(repo.name, diff)

    def pushables(self, remote):
        for name in self.repositories:
            repo = self.db.repositories[name]
            commits = repo.pushables(remote)
            if len(commits):
                yield Box(repo.name, "\n".join(commits))

    def untracked(self):
        for name in self.repositories:
            repo = self.db.repositories[name]
            files = repo.get_untracked_files()
            if len(files):
                yield Box(repo.name, "\n".join(files))

    def dirty(self):
        for name in self.repositories:
            repo = self.db.repositories[name]
            files = repo.get_dirty_files()
            if len(files):
                yield Box(repo.name, "\n".join(files))

    def fetch(self):
        for name in self.repositories:
            repo = self.db.repositories[name]
            res = repo.fetch()
            if len(res):
                yield Box(repo.name, res)

    def status(self):
        for name in self.repositories:
            repo = self.db.repositories[name]
            yield Box(repo.name, repo.status())

    def cmd(self, command):
        for repo_name in self.repositories:
            repo = self.db.repositories[repo_name]
            res = repo.cmd(command)
            if len(res):
                yield Box(repo.name, res)

    def __repr__(self):
        return "<Project: %s>" % self.__dict__()

    def __dict__(self):
        return dict(
            name = self.name,
            repositories = sorted(self.repositories),
            description = self.description,
        )




from colors import Colors

class BoxRenderer(object):
    def __init__(self, width = 100, decor = "-"):
        self.width = width
        self.decor = decor

    def render(self, box):
        padding = 10

        res  = Colors.yellow + self.decor * padding
        res += Colors.green + " %s " % box.caption
        res += Colors.yellow + self.decor * (self.width - len(box.caption) - 2 - padding)
        res += Colors.default + "\n" + box.content

        return res




import subprocess
from .repository import Repository

class GitRepository(Repository):
    type = 'git'

    def diff(self):
        return self.cmd("git --no-pager diff")

    def pushables(self, remote):
        return self.list_cmd("git --no-pager log --oneline %s..HEAD" % remote)

    def get_new_commits(self):
        return self.list_cmd("git --no-pager log --oneline HEAD..origin/master")

    def fetch(self):
        return self.cmd("git fetch")

    def status(self):
        return self.cmd("git status")

    def get_dirty_files(self):
        files = []
        for file in self.list_cmd("git status --short"):
            if file.find('?') == -1:
                files.append(file)
        return files

    def get_untracked_files(self):
        return self.list_cmd("git ls-files --others --exclude-standard")




class Storage(dict):

    def __getattr__(self, key):
        return self[key]

    def __setattr__(self, key, value):
        self[key] = value

    def __hasattr__(self, key):
        return key in self

class CLIArgumentsTreeParser(object):

    def __init__(self, config, root_name, parser):
        self.parser = parser
        self.config = config
        self.root_name = root_name

    def __build(self, parser, node, node_name):
        subparsers = parser.add_subparsers(dest = node_name)

        for item in node:
            subparser = subparsers.add_parser(item['name'], **item['desc'])

            if 'arguments' in item:
                for arg in item['arguments']:
                    name = arg['arg_name']
                    del arg['arg_name']
                    if isinstance(name, list):
                        subparser.add_argument(*name, **arg)
                    else:
                        subparser.add_argument(name, **arg)

            if 'subcommands' in item:
                self.__build(subparser, item['subcommands'], item['name'])

    def build(self):
        self.__build(self.parser, self.config, self.root_name)

    def parse(self):
        self.raw_data = vars(self.parser.parse_args())

        self.data = Storage()

        self.structuring(self.root_name, self.data)

        return self.data

    def structuring(self, name, result, nodes = []):
        if name in self.raw_data:
            nodes.append(name)
            result.name = name
            if len(self.raw_data) > len(nodes):
                result.sub = Storage()
                self.structuring(self.raw_data[name], result.sub, nodes)
            else:
                result.sub = Storage(name = self.raw_data[name])
        else:
            result.name = name
            result.args = Storage()
            for dname in self.raw_data:
                if dname not in nodes:
                    result.args[dname] = self.raw_data[dname]




import os
import pty
from subprocess import Popen, PIPE
from abc import ABCMeta, abstractmethod

def num_bytes_readable(fd):
    import array
    import fcntl
    import termios
    buf = array.array('i', [0])
    if fcntl.ioctl(fd, termios.FIONREAD, buf, 1) == -1:
        raise Exception("We really should have had data")
    return buf[0]

class Repository(object):
    __metaclass__ = ABCMeta
    type = 'notimplemented'

    def __init__(self, path, name):
        self.path = path
        self.name = name

    def list_cmd(self, command):
        return self.cmd(command).split("\n")[:-1]

    def cmd(self, command):
        # pty for colored output
        master, slave = pty.openpty()
        p = Popen(command, cwd = self.path, shell = True, stdout = slave, stderr = slave)
        p.communicate()
        return os.read(master, num_bytes_readable(master))

    @abstractmethod
    def diff(self):
        pass

    @abstractmethod
    def fetch(self):
        pass

    @abstractmethod
    def pushables(self, remote):
        pass

    @abstractmethod
    def status(self):
        pass

    @abstractmethod
    def get_untracked_files(self):
        pass

    @abstractmethod
    def get_dirty_files(self):
        pass

    def __repr__(self):
        return "<Repository: %s>" % self.__dict__()

    def __dict__(self):
        return dict(
            path = self.path,
            name = self.name,
            type = self.type,
        )




import os
from prettytable import PrettyTable
from functools import partial
from logging import getLogger

from .project import Project
from .box_renderer import BoxRenderer
from .git_repository import GitRepository
from .exceptions import RepositoryException, ProjectException

logger = getLogger(__name__)

class RepositoryFactory(object):
    def __init__(self):
        self.types = {}
        self.types[GitRepository.type] = GitRepository

    def create(self, path, type, name):
        return self.types[type](path, name)

class ProjectModifyRepositoryCommand(object):
    def __init__(self, project):
        self.project = project

    def add(self, names):
        self.project.repositories += list(set(names))
        logger.info("Repo added to project '%s'" % self.project.name)

    def clear(self):
        self.project.repositories = []
        logger.info("Repositories cleared from project '%s'" % self.project.name)

    def remove(self, names):
        self.project.repositories = list(set(self.project.repositories) - set(names))
        logger.info("Repo removed from project '%s'" % self.project.name)

class ProjectModifyCommand(object):
    def __init__(self, project):
        self.project = project

    def repository(self):
        return ProjectModifyRepositoryCommand(self.project)

    def description(self, data):
        self.project.description = data
        logger.info("Description modified")

class DyanmicProjectModifyCommand(object):
    def __init__(self, vcp):
        self.vcp = vcp

    def __hasattr__(self, name):
        return name in self.vcp.projects

    def __getattr__(self, name):
        return partial(ProjectModifyCommand, self.vcp.projects[name])

class ProjectCommand(object):
    def __init__(self, vcp):
        self.vcp = vcp

    def create(self, name, description, repositories):
        if name in self.vcp.projects:
            raise ProjectException("Project '%s' is already exists" % name)
        self.vcp.projects[name] = Project(name, description, repositories)
        logger.info("Project '%s' created" % name)

    def remove(self, name):
        del self.vcp.projects[name]
        logger.info("Project '%s' removed" % name)

    def modify(self):
        return DyanmicProjectModifyCommand(self.vcp)

    def show(self, name):
        project = self.vcp.projects[name]

        logger.info("Repositories of '%s'(%s) project:" % (project.name, project.description))

        table = PrettyTable(["Type", "Name", "Path"])
        table.align = 'l'
        for repo_name in project.repositories:
            repo = self.vcp.repositories[repo_name]
            table.add_row([repo.type, repo.name, repo.path])
        logger.info(table)

    def list(self):
        logger.info("Known projects:")
        table = PrettyTable(["Name", "Description", "Repo cnt"])
        table.align = 'l'
        table.align["Repo cnt"] = "r"
        for project in self.vcp.projects.values():
            table.add_row([project.name, project.description, len(project.repositories)])

        logger.info(table)

class RepositoryCommand(object):
    def __init__(self, vcp):
        self.vcp = vcp

    def cmd(self, name, command):
        logger.info(self.vcp.repositories[name].cmd(command))

    def list(self):
        logger.info("Known repositories")

        table = PrettyTable(["Type", "Name", "Path"])
        table.align = 'l'
        for repo_name in self.vcp.repositories:
            repo = self.vcp.repositories[repo_name]
            table.add_row([repo.type, repo.name, repo.path])
        logger.info(str(table))

    def remove(self, name):
        del self.vcp.repositories[name]
        for prj_name in self.vcp.projects:
            project = self.vcp.projects[prj_name]
            if name in project.repositories:
                project.repositories.remove(name)
                logger.info("Repository '%s' removed from project '%s'" % (name, prj_name))
        logger.info("Repository '%s' removed" % name)

    def create(self, path, type, name, add_to = None):
        path = os.path.abspath(path)
        if not os.path.isdir(path):
            raise RepositoryException("Path '%s' is not exists" % path)
        if name is None:
            name = path.split(os.path.sep)[-1:][0]

        if name in self.vcp.repositories:
            raise ProjectException("Repository '%s' is already exists" % name)

        self.vcp.repositories[name] = self.vcp.repo_factory.create(path, type, name)

        msg = "Repository '%s' created" % name

        if add_to is not None:
            add_to = set(add_to)
            for prj_name in add_to:
                self.vcp.projects[prj_name].repositories.append(name)
            msg += " and added to project: " + ', '.join(add_to)

        logger.info(msg)

class VCP(object):

    def __init__(self, config):
        self.config = config
        self.projects = {}
        self.repositories = {}
        self.box_renderer = BoxRenderer()
        self.repo_factory = RepositoryFactory()

        if 'projects' in config:
            for name in config['projects']:
                project = Project(**config['projects'][name])
                project.db = self
                self.projects[name] = project

        if 'repositories' in config:
            for name in config['repositories']:
                self.repositories[name] = self.repo_factory.create(**config['repositories'][name])

    def news(self, name, fromcache):
        project = self.projects[name]

        for box in project.news(fromcache):
            logger.info(self.box_renderer.render(box))

    def diff(self, name):
        project = self.projects[name]
        for box in project.diff():
            logger.info(self.box_renderer.render(box))

    def pushables(self, name, remote):
        project = self.projects[name]

        for box in project.pushables(remote):
            logger.info(self.box_renderer.render(box))

    def untracked(self, name):
        project = self.projects[name]

        for box in project.untracked():
            logger.info(self.box_renderer.render(box))

    def dirty(self, name):
        project = self.projects[name]

        for box in project.dirty():
            logger.info(self.box_renderer.render(box))

    def cmd(self, name, command):
        project = self.projects[name]

        for box in project.cmd(command):
            logger.info(self.box_renderer.render(box))

    def fetch(self, name):
        project = self.projects[name]

        for box in project.fetch():
            logger.info(self.box_renderer.render(box))

    def status(self, name):
        project = self.projects[name]

        for box in project.status():
            logger.info(self.box_renderer.render(box))

    def repository(self):
        return RepositoryCommand(self)

    def project(self):
        return ProjectCommand(self)

    def get_data(self):
        return dict(
            projects = self.projects,
            repositories = self.repositories,
        )



from setuptools import setup, find_packages
from setuptools.command.install import install
import subprocess

setup(
    name = "vcp",
    version = "1.2.0",
    description = "Version Control Projects - organize repositories",
    author = 'Lajos Santa',
    author_email = 'santa.lajos@coldline.hu',
    url = 'https://github.com/voidpp/vcp.git',
    license = 'MIT',
    install_requires = [
        "prettytable==0.7.2",
        "argcomplete==1.0.0",
    ],
    packages = find_packages(),
    scripts = [
        'bin/vcp',
    ],
)
